#include "lis.h"

#include "random_walk.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

/*
 * Parses the program arguments:
 * -l label_file
 * -k number_of_labels
 * -i input_image
 * -o output_image
 */
static void ParseOpt (int argc, char **argv, char **input_img_file,
        char **output_img_file, char **label_file, int *num_labels) {
    int c;
    while ((c = getopt (argc, argv, "l:k:i:o:")) != -1) {
        switch (c)
        {
            case 'l':
                *label_file = optarg;
                break;
            case 'k':
                *num_labels = (int) atoi (optarg);
                break;
            case 'i':
                *input_img_file = optarg;
                break;
            case 'o':
                *output_img_file = optarg;
                break;
            default:
                exit (EXIT_FAILURE);
        }
    }
}

/*
 * Validates the program parameters.
 * Checks if it can read the input image and the labels file,
 * if it can write to the output image, and if the number of labels
 * is at least 2.
 */
static void ValidateOpt (char *input_img_file, char *output_img_file,
                         char *label_file, int num_labels) {
    if (label_file == NULL) {
        fprintf (stderr, "Must pass input label file as parameter\n");
        exit (EXIT_FAILURE);
    }
    if (access (label_file, R_OK)) {
        perror ("Invalid input label file");
        exit (EXIT_FAILURE);
    }
    if (input_img_file == NULL) {
        fprintf (stderr, "Must pass input image file as parameter\n");
        exit (EXIT_FAILURE);
    }
    if (access (input_img_file, R_OK)) {
        perror ("Invalid input image file");
        exit (EXIT_FAILURE);
    }
    if (output_img_file == NULL) {
        fprintf (stderr, "Must pass output label file as parameter\n");
        exit (EXIT_FAILURE);
    }
    if (access (output_img_file, W_OK)) {
        if (errno != ENOENT) { /* ENOENT means file does not exist, but that's ok */
            perror ("Invalid output image file");
            exit (EXIT_FAILURE);
        }
    }
    if (num_labels <= 1) {
        fprintf (stderr, "Invalid number of labels: Must be at least 2\n");
        exit (EXIT_FAILURE);
    }
}

/* 
 * The main function!
 * Parses arguments and executes the Random Walk algorithm on the input image.
 */
int main(int argc, char **argv) {
    int exit_value = 0;
    char *label_file = NULL, *input_img_file = NULL, *output_img_file = NULL;
    int num_labels = 0;
    RandomWalk rw;

    ParseOpt (argc, argv, &input_img_file, &output_img_file,
              &label_file, &num_labels);
    ValidateOpt (input_img_file, output_img_file, label_file, num_labels);

    rw = RANDOMWALKinit (input_img_file, output_img_file, label_file, num_labels);
    if ((exit_value = RANDOMWALKrun (rw))) {
        fprintf (stderr, "Error performing segmentation\n");
    }
    RANDOMWALKdestroy (&rw);

    return exit_value;
}

