#include "random_walk.h"
#include "temp_sparse_matrix.h"

#include <stdio.h>

/*
 * Initializes the Random Walk algorithm (doesn't execute it), returning the RandomWalk file that must be passed to the other functions.
 * input_img_file: file name of the input pgm image to segment
 * output_img_file: file name of the output pgm file, with the labels indexes assigned by the segmentation
 * label_file: pgm file with the labels indexes from 1 to n. Pixels that will be assigned to one of the labels must have the value 0.
 * num_labels: number of labels int the label_file (the 'n' mentioned above).
 */
RandomWalk RANDOMWALKinit (char *input_img_file, char *output_img_file,
        char *label_file, int num_labels) {
    RandomWalk rw = malloc (sizeof (*rw));
    if (rw == NULL) {
        fprintf (stderr, "Error allocating memory!\n");
        return NULL;
    }
    rw->lin_sys = malloc (sizeof (*rw->lin_sys));
    if (rw->lin_sys == NULL) {
        fprintf (stderr, "Error allocating memory!\n");
        free (rw);
        return NULL;
    }
    LIS_INT argc = 1;
    char **argv = 0;
    int i;
    lis_initialize (&argc, &argv);
    rw->input = ReadImage (input_img_file);
    rw->output = CreateImage (rw->input->ncols, rw->input->nrows);
    for (i = 0; i < rw->output->ncols * rw->output->nrows; ++i) {
        rw->output->val[i] = 0;
    }
    rw->labels = ReadImage (label_file);
    rw->num_labels = num_labels;
    rw->output_img_file = output_img_file;
    rw->adj = Circular (1.0); /* 4-neighborhood */
    rw->num_unlabeled = 0;
    rw->lin_sys->B = NULL;
    rw->last_x = NULL;
    rw->max_prob = NULL;
    return rw;
}

/* Frees up the memory on the rw struct */
void RANDOMWALKdestroy (RandomWalk *rw) {
    DestroyImage (&((*rw)->labels));
    DestroyImage (&((*rw)->output));
    free ((*rw)->lin_sys);
    free (*rw);
    *rw = NULL;
    lis_finalize ();
}

/* Process the labels_ image to index the image pixels */
static int IndexPixels (Image *labels_, int *num_labels_, int *num_unlabeled_) {
    int index = 0, num_labels = 0;
    int p;
    for (p = 0; p < labels_->ncols * labels_->nrows; ++p) {
        if (labels_->val[p] == 0) { /* Not labeled */
            labels_->val[p] = index++;
        } else { /* Labels will be the negative values */
            if (labels_->val[p] > num_labels)
                num_labels = labels_->val[p];
            labels_->val[p] = -labels_->val[p];
        }
    }
    *num_unlabeled_ = index;
    if (num_labels_ == 0)
        *num_labels_ = num_labels;
    if (*num_labels_ != num_labels) {
        fprintf (stderr, "Specified %d labels, but found %d label(s) "
                         "on label file\n", *num_labels_, num_labels);
        return 1;
    }
    return 0;
}

/* 
 * Returns the weight between pixels with indexes p and q
 * of the input image
 */
static double GetWeight (Image *input, int p, int q) {
    /*return 256 - abs (input_->val[p] - input_->val[q]);*/
    double e = 2.71828182846;
    double diff, normalized_squared_diff;
    double weight;
    diff = input->val[p] - input->val[q];
    normalized_squared_diff = (diff * diff) / (255 * 255);
    weight = pow (e, -200 * normalized_squared_diff);
    if (weight == 0) {
        fprintf (stderr, "Error: weight is 0 for values %d and %d\n", input->val[p], input->val[q]);
        /*fprintf (stderr, "diff = %lf, normalized_squared_diff = %lf\n", diff, normalized_squared_diff);*/
    }
    return weight;
}

/*
 * Creates a temporary matrix to store the 'b' vectors of the linear equations.
 */
static double ** CreateBTemp (int num_labels, int num_unlabeled) {
    int i, j;
    double **B_tmp = NULL;
    B_tmp = (double **) malloc (sizeof(*B_tmp) * num_labels);
    if (B_tmp == NULL) {
        fprintf (stderr, "Error allocating B_tmp matrix\n");
        return NULL;
    }
    for (i = 0; i < num_labels; ++i) {
        B_tmp[i] = (double *) malloc (sizeof(double) * num_unlabeled);
        if (B_tmp[i] == NULL) {
            fprintf (stderr, "Error allocating B_tmp matrix\n");
            return NULL;
        }
        for (j = 0; j < num_unlabeled; ++j)
            B_tmp[i][j] = 0.0;
    }
    return B_tmp;
}

/*
 * Destroys the temporary B matrix created by CreateBTemp,
 * freeing its memory.
 */
static void DestroyBTemp (double ***B_tmp, int num_labels) {
    int i;
    for (i = 0; i < num_labels; ++i) {
        free ((*B_tmp)[i]);
        (*B_tmp)[i] = NULL;
    }
    free (*B_tmp);
    *B_tmp = NULL;
}

#define UNLABELED(p) (p >= 0)

/*
 * Fills the A and B matrixes of the linear system
 * (B is the matrix with all the b vectors of the
 * linear systems that will be solved)
 */
static int FillAAndB (RandomWalk rw) {
    int p, x, y, p_index;
    int q, adj_x, adj_y, q_index;
    int i, j;
    double weight;
    double **B_tmp = NULL;
    TmpMatrix T = CreateTempSparseMatrix (rw->num_unlabeled, rw->adj->n); /* second parameter is num of neighbours + 1 */

    B_tmp = CreateBTemp (rw->num_labels, rw->num_unlabeled);
    if (B_tmp == NULL)
        return 1;

    /* Traversing all the unlabeled pixels */
    for (y = p = 0; y < rw->input->nrows; ++y) {
        for (x = 0; x < rw->input->ncols; ++x, ++p) {
            p_index = rw->labels->val[p];
            if (UNLABELED(p_index)) {
                SetMatrixValue (T, p_index, p_index, 0);
                /* Traversing all the neighbouring pixels of the
                 * current pixel */
                for(i = 0; i < rw->adj->n; ++i){
                    adj_x = x + rw->adj->dx[i];
                    adj_y = y + rw->adj->dy[i];
                    if ((adj_x == x) && (adj_y == y))
                        continue;

                    if (ValidPixel (rw->labels, adj_x, adj_y)){
                        q = adj_x + adj_y * rw->labels->ncols;
                        q_index = rw->labels->val[q];
                        /* Filling the temporary T and B_tmp matrixes */
                        if (UNLABELED(q_index)) {
                            if (q_index > p_index) {
                                weight = GetWeight (rw->input, p, q);
                                SetMatrixValue (T, p_index, q_index, -weight);
                                AddMatrixValue (T, p_index, p_index, weight);
                            } else {
                                /* We can use the previously calculated weight, instead of calculating it again */
                                CopyNewMatrixValue (T, q_index, p_index, p_index, q_index, &weight);
                                AddMatrixValue (T, p_index, p_index, -weight);
                            }
                        } else {
                            q_index = -q_index - 1;
                            weight = GetWeight (rw->input, p, q);
                            AddMatrixValue (T, p_index, p_index, weight);
                            B_tmp[q_index][p_index] += weight;
                        }
                    }
                }
            }
        }
    }
    DestroyImage (&(rw->input));
    /* Passing the B_tmp values to the final B matrix */
    for (i = 0; i < rw->num_labels; ++i)
        for (j = 0; j < rw->num_unlabeled; ++j)
            lis_vector_set_value (LIS_INS_VALUE, j, B_tmp[i][j], rw->lin_sys->B[i]);
    DestroyBTemp (&B_tmp, rw->num_labels);
    /* Transforming the T matrix to the correct format,
     * assignint it to the A matrix */
    rw->lin_sys->A = GetLisMatrix (T);
    DestroyTempSparseMatrix (&T);
    return 0;
}

/*
 * Creates the linear systems to be solved,
 * allocating memory and filling the matrixes values.
 */
static int CreateLinearSystem (RandomWalk rw) {
    LIS_INT    A_n = rw->num_unlabeled;
    LIS_INT    x_n = A_n;
    LIS_INT    b_n = A_n;
    int l;

    rw->lin_sys->B = (LIS_VECTOR *) malloc (sizeof(*(rw->lin_sys->B)) * rw->num_labels);
    if (rw->lin_sys->B == NULL) {
        fprintf (stderr, "Error allocating B matrix\n");
        return 1;
    }
    for (l = 0; l < rw->num_labels; ++l) {
        lis_vector_create (0, rw->lin_sys->B + l);
        lis_vector_set_size (rw->lin_sys->B[l], 0, b_n);
    }

    lis_vector_create (0, &(rw->lin_sys->x));
    lis_vector_set_size (rw->lin_sys->x, 0, x_n);
    lis_vector_set_all (0, rw->lin_sys->x);

    rw->last_x = (double *) malloc (sizeof(*(rw->last_x)) * x_n);
    if (rw->last_x == NULL) {
        fprintf (stderr, "Error allocating last_x vector\n");
        return 1;
    }
    for (l = 0; l < x_n; ++l)
        rw->last_x[l] = 1.;

    rw->max_prob = (double *) malloc (sizeof(*(rw->max_prob)) * x_n);
    if (rw->max_prob == NULL) {
        fprintf (stderr, "Error allocating max_prob vector\n");
        return 1;
    }
    for (l = 0; l < x_n; ++l)
        rw->max_prob[l] = 0.;

    if (FillAAndB (rw))
        return 2;

    return 0;
}

/*
 * Frees up the memory allocated by CreateLinearSystem
 */
static int DestroyLinearSystem (LIS_MATRIX A, LIS_VECTOR x, LIS_VECTOR **B, double **last_x, double **max_prob, int num_labels) {
    int l;
    for (l = 0; l < num_labels; ++l) {
        lis_vector_destroy ((*B)[l]);
    }
    free(*B);
    *B = NULL;
    lis_vector_destroy (x);
    lis_matrix_destroy (A);
    free (*last_x);
    *last_x = NULL;
    free (*max_prob);
    *max_prob = NULL;
    return 0;
}

/*
 * Returns a sequenced image file name for the labels probabilities.
 * This file name is based on the string output_img_file_ and the number label_num.
 */
static char * CreateLabelImageFileName (char *output_img_file_, int label_num) {
    char *img_file_name = NULL;
    int img_file_name_size, n, m;
    char buffer[50];
    m = strlen(output_img_file_);
    if (m >= 4 && strcmp (output_img_file_ + m - 4, ".pgm") == 0)
        m -= 4;
    n = sprintf (buffer, "_prob_label_%d.pgm", label_num);
    img_file_name_size = m + n + 1;
    img_file_name = (char *) malloc (sizeof(char) * img_file_name_size);
    if (img_file_name == NULL) {
        fprintf (stderr, 
                "Error allocating memory for label image of the label %d\n",
                label_num);
        return NULL;
    }
    strncpy (img_file_name, output_img_file_, m);
    img_file_name[m] = '\0';
    strncat (img_file_name, buffer, n);
    return img_file_name;
}

/*
 * Writes the label probabilities image file, and also the final
 * segmentation if it is the last label being written.
 */
static void WriteLabelImage (char *output_img_file, Image *output, Image *labels, int label_num, LIS_VECTOR x, double *last_x, double *max_prob, int is_last_label) {
    int k, pixel_index;
    double value;
    Image *img = NULL;
    char *img_file_name = NULL;
    img_file_name = CreateLabelImageFileName (output_img_file, label_num);
    img = CreateImage (labels->ncols, labels->nrows);
    if (img_file_name == NULL || img == NULL) {
        fprintf (stderr,
                "Error creating label image for label %d\n", label_num);
        return;
    }
    for (k = 0; k < labels->ncols * labels->nrows; ++k) {
        if (labels->val[k] >= 0) {
            pixel_index = labels->val[k];
            if (!is_last_label) {
                lis_vector_get_value (x, pixel_index, &value);
                last_x[labels->val[k]] -= value;
            } else {
                value = last_x[pixel_index];
            }
            if (value > max_prob[pixel_index]) {
                max_prob[pixel_index] = value;
                output->val[k] = label_num;
            }
            img->val[k] = (int) (value * 255.0 + 0.5);
            if (img->val[k] < 0) {
                fprintf (stderr, "Found negative probability!\n");
                img->val[k] = 0;
            } else if (img->val[k] > 255) {
                fprintf (stderr, "Found probability bigger than 1!\n");
                img->val[k] = 255;
            }
        } else {
            /* This pixel is already labeled by the input */
            if (labels->val[k] == -label_num) {
                img->val[k] = 255;
                output->val[k] = label_num;
            } else {
                img->val[k] = 0;
            }
        }
    }
    WriteImage (img, img_file_name);
    DestroyImage (&img);
    free (img_file_name);
    img_file_name = NULL;
    if (is_last_label) {
        WriteImage (output, output_img_file);
    }
}

/*
 * Solves the linear system and writes the results
 * to label probabilities images files.
 * Also writes the final segmentation, with the assigned label indexes,
 * on the output_img_file.
 */
static void Solve (RandomWalk rw) {
    LIS_SOLVER solver;
    LIS_INT iter;
    int l;

    lis_solver_create(&solver);
    lis_solver_set_optionC(solver);
    lis_solver_set_option("-i cg -p ssor", solver);
    lis_solver_set_option("-tol 1.0e-12 -maxiter 100000", solver);
    for (l = 0; l < rw->num_labels - 1; ++l) {
        lis_solve (rw->lin_sys->A, rw->lin_sys->B[l], rw->lin_sys->x, solver);
        lis_solver_get_iters (solver, &iter);
        printf ("Number of iterations to solve linear system: %d\n", iter);
        WriteLabelImage (rw->output_img_file, rw->output, rw->labels, l + 1, rw->lin_sys->x, rw->last_x, rw->max_prob, 0);
    }
    WriteLabelImage (rw->output_img_file, rw->output, rw->labels, l + 1, rw->lin_sys->x, rw->last_x, rw->max_prob, 1);
    lis_solver_destroy(solver);
}

/* Executes the Random Walks algorithm with the information stored on rw */
int RANDOMWALKrun (RandomWalk rw) {
    if (rw->input->ncols != rw->labels->ncols || rw->input->nrows != rw->labels->nrows) {
        fprintf (stderr, "Input image size and label size differ!\n");
        return 1;
    }
    if (IndexPixels (rw->labels, &(rw->num_labels), &(rw->num_unlabeled)))
        return 2;
    if (CreateLinearSystem (rw))
        return 3;
    printf ("Unlabeled pixels: %d\n", rw->num_unlabeled);
    Solve (rw);
    DestroyLinearSystem (rw->lin_sys->A, rw->lin_sys->x, &(rw->lin_sys->B), &(rw->last_x), &(rw->max_prob), rw->num_labels);
    return 0;
}
