#include "temp_sparse_matrix.h"

#include <stdlib.h>

TmpMatrix CreateTempSparseMatrix (int nrows, int maxcols) {
    int k;
    TmpMatrix A = (TmpMatrix) malloc (sizeof(*A));
    if (A != NULL) {
        A->nrows = nrows;
        A->maxcols = maxcols;
        A->nnz = 0;
        A->row_size = maxcols + 1; /* add one space to sentinel */
        A->val = (double *) malloc (sizeof(double) * nrows * A->row_size);
        if (A->val == NULL)
            return NULL;
        A->col_num = (int *) malloc (sizeof(int) * nrows * A->row_size);
        if (A->col_num == NULL)
            return NULL;
        for (k = 0; k < nrows; ++k) {
            A->col_num[k * A->row_size] = -1; /* Sentinel value */
        }
    }
    return A;
}

void DestroyTempSparseMatrix (TmpMatrix *A) {
    if (A == NULL)
        return;
    if (*A == NULL)
        return;
    free ((*A)->val);
    (*A)->val = NULL;
    free ((*A)->col_num);
    (*A)->col_num = NULL;
    free (*A);
    *A = NULL;
}

int SetMatrixValue (TmpMatrix A, int i, int j, double value) {
    int k;
    if (A == NULL)
        return 1;
    k = i * A->row_size;
    while (1) {
        if (A->col_num[k] == -1) {
            A->val[k] = value;
            A->col_num[k] = j;
            A->col_num[k + 1] = -1;
            A->nnz++;
            return 0;
        }
        ++k;
    }
}

int AddMatrixValue (TmpMatrix A, int i, int j, double value) {
    int k;
    if (A == NULL)
        return 1;
    for (k = i * A->row_size; A->col_num[k] != j; ++k);
    A->val[k] += value;
    return 0;
}

int CopyNewMatrixValue (TmpMatrix A, int src_i, int src_j,
                     int dest_i, int dest_j, double *value) {
    int k, l;
    if (A == NULL)
        return 1;
    for (k = src_i * A->row_size; A->col_num[k] != src_j; ++k);
    for (l = dest_i * A->row_size; A->col_num[l] != -1; ++l);
    A->val[l] = A->val[k];
    A->col_num[l] = dest_j;
    A->col_num[l + 1] = -1;
    A->nnz++;
    if (value != NULL)
        *value = A->val[l];
    return 0;
}

static void SortRowValues (TmpMatrix A) {
    /* sort row values according to column index (insertion sort) */
    int m, i, j, start;
    double aux, aux_col;
    for (m = 0; m < A->nrows; ++m) {
        start = m * A->row_size;
        for (i = start; A->col_num[i] != -1; ++i) {
            aux = A->val[i];
            aux_col = A->col_num[i];
            for (j = i; j > start && A->col_num[j - 1] > aux_col; --j) {
                A->val[j] = A->val[j - 1];
                A->col_num[j] = A->col_num[j - 1];
            }
            A->val[j] = aux;
            A->col_num[j] = aux_col;
        }
    }
}

LIS_MATRIX GetLisMatrix (TmpMatrix A) {
    int i, j, ptr_idx, value_idx;
    LIS_INT    n, nnz;
    LIS_INT    *ptr, *index;
    LIS_SCALAR *value;
    LIS_MATRIX B;
    n = A->nrows;
    nnz = A->nnz;
    lis_matrix_malloc_csr (n, nnz, &ptr, &index, &value);
    lis_matrix_create (0, &B);
    lis_matrix_set_size (B, 0, n);

    SortRowValues (A);

    ptr_idx = 0;
    value_idx = 0;
    for (i = 0; i < A->nrows; ++i) {
        ptr[ptr_idx++] = value_idx;
        for (j = i * A->row_size; A->col_num[j] != -1; ++j) {
            if (A->val[j] != 0) {
                index[value_idx] = A->col_num[j];
                value[value_idx] = A->val[j];
                value_idx++;
            }
        }
    } 
    ptr[n] = nnz;

    lis_matrix_set_csr (nnz, ptr, index, value, B);
    lis_matrix_assemble (B);
    return B;
}

