#include "random_walk.h"
#include "temp_sparse_matrix.h"

#include <stdio.h>

Scene* ReadMarkers (char *markers_file_name, int xsize, int ysize, int zsize) {
    int i, n;
    int x, y, z, id, label_num;
    FILE *markers = NULL;

    Scene *labels = CreateScene (xsize, ysize, zsize);
    for (i = 0; i < labels->n; ++i){
        labels->data[i] = 0;
    }
    labels->maxval = 0;

    markers = fopen (markers_file_name, "r");
    if (markers == NULL) {
        fprintf (stderr, "Error readind markers file.\n");
	return NULL;
    }
    fscanf (markers, "%d", &n);
    for (i = 0; i < n; ++i) {
        fscanf (markers, "%d %d %d %d %d", &x, &y, &z, &id, &label_num);
	if (ValidVoxel (labels, x, y, z)) {
	    SetVoxel (labels, VoxelAddress (labels, x, y, z), label_num + 1);
	} else {
	    fprintf (stderr, "Voxel with invalid coordinates (%d, %d, %d) in labels file.\n", x, y, z);
	}
    }
    fclose (markers);
    return labels;
}

/*
 * Initializes the Random Walk algorithm (doesn't execute it), returning
 * the RandomWalk file that must be passed to the other functions.
 * input_scene_file: file name of the input scn scene to segment
 * output_scene_file: file name of the output scn file,
 *                    with the labels indexes assigned by the segmentation
 * label_file: scn file with the labels indexes from 1 to n. Voxels that
 *             will be assigned to one of the labels must have the value 0.
 * num_labels: number of labels int the label_file (the 'n' mentioned above).
 */
RandomWalk RANDOMWALKinit (char *input_scene_file, char *output_scene_file,
        char *label_file, int num_labels) {
    RandomWalk rw = malloc (sizeof (*rw));
    if (rw == NULL) {
        fprintf (stderr, "Error allocating memory!\n");
        return NULL;
    }
    rw->lin_sys = malloc (sizeof (*rw->lin_sys));
    if (rw->lin_sys == NULL) {
        fprintf (stderr, "Error allocating memory!\n");
        free (rw);
        return NULL;
    }
    LIS_INT argc = 1;
    char **argv = 0;
    int i;
    lis_initialize (&argc, &argv);
    rw->input = ReadScene (input_scene_file);
    rw->maxval = SceneImax (rw->input);
    rw->output = CreateScene (rw->input->xsize, rw->input->ysize, rw->input->zsize);
    for (i = 0; i < rw->output->n; ++i) {
        rw->output->data[i] = 0;
    }
    rw->output->maxval = 0;
    rw->labels = ReadMarkers (label_file, rw->input->xsize, rw->input->ysize, rw->input->zsize);
    rw->num_labels = num_labels;
    rw->output_scene_file = output_scene_file;
    rw->adj = Spheric (1.0); /* 6-neighborhood */
    rw->num_unlabeled = 0;
    rw->lin_sys->B = NULL;
    rw->last_x = NULL;
    rw->max_prob = NULL;
    return rw;
}

/* Frees up the memory on the rw struct */
void RANDOMWALKdestroy (RandomWalk *rw) {
    DestroyScene (&((*rw)->labels));
    DestroyScene (&((*rw)->output));
    free ((*rw)->lin_sys);
    free (*rw);
    *rw = NULL;
    lis_finalize ();
}

/* Process the labels_ scene to index the scene voxels */
static int IndexVoxels (Scene *labels_, int *num_labels_, int *num_unlabeled_) {
    int index = 0, num_labels = 0;
    int p;
    for (p = 0; p < labels_->n; ++p) {
        if (labels_->data[p] == 0) { /* Not labeled */
            labels_->data[p] = index++;
        } else { /* Labels will be the negative values */
            if (labels_->data[p] > num_labels)
                num_labels = labels_->data[p];
            labels_->data[p] = -labels_->data[p];
        }
    }
    *num_unlabeled_ = index;
    if (*num_labels_ == 0)
        *num_labels_ = num_labels;
    if (*num_labels_ != num_labels) {
        fprintf (stderr, "Specified %d labels, but found %d label(s) "
                         "on label file\n", *num_labels_, num_labels);
        return 1;
    }
    return 0;
}

/* 
 * Returns the weight between voxels with indexes p and q
 * of the input scene
 */
static double GetWeight (Scene *input, int p, int q) {
    /*return pow(input->maxval + 1 - abs (input->data[p] - input->data[q]), 7);*/
    double e = 2.71828182846;
    double diff, normalized_squared_diff;
    double weight;
    diff = input->data[p] - input->data[q];
    normalized_squared_diff = (diff * diff) / (input->maxval * input->maxval);
    weight = pow (e, -200 * normalized_squared_diff);
    if (weight == 0) {
        fprintf (stderr, "Error: weight is 0 for values %d and %d\n", input->data[p], input->data[q]);
        fprintf (stderr, "diff = %lf, normalized_squared_diff = %lf\n", diff, normalized_squared_diff);
    }
    return weight;
}

/*
 * Creates a temporary matrix to store the 'b' vectors of the linear equations.
 */
static double ** CreateBTemp (int num_labels, int num_unlabeled) {
    int i, j;
    double **B_tmp = NULL;
    B_tmp = (double **) malloc (sizeof(*B_tmp) * num_labels);
    if (B_tmp == NULL) {
        fprintf (stderr, "Error allocating B_tmp matrix\n");
        return NULL;
    }
    for (i = 0; i < num_labels; ++i) {
        B_tmp[i] = (double *) malloc (sizeof(double) * num_unlabeled);
        if (B_tmp[i] == NULL) {
            fprintf (stderr, "Error allocating B_tmp matrix\n");
            return NULL;
        }
        for (j = 0; j < num_unlabeled; ++j)
            B_tmp[i][j] = 0.0;
    }
    return B_tmp;
}

/*
 * Destroys the temporary B matrix created by CreateBTemp,
 * freeing its memory.
 */
static void DestroyBTemp (double ***B_tmp, int num_labels) {
    int i;
    for (i = 0; i < num_labels; ++i) {
        free ((*B_tmp)[i]);
        (*B_tmp)[i] = NULL;
    }
    free (*B_tmp);
    *B_tmp = NULL;
}

#define UNLABELED(p) (p >= 0)

/*
 * Fills the A and B matrixes of the linear system
 * (B is the matrix with all the b vectors of the
 * linear systems that will be solved)
 */
static int FillAAndB (RandomWalk rw) {
    int p, x, y, z, p_index;
    int q, adj_x, adj_y, adj_z, q_index;
    int i, j;
    double weight;
    double **B_tmp = NULL;
    TmpMatrix T = CreateTempSparseMatrix (rw->num_unlabeled, rw->adj->n); /* second parameter is num of neighbours + 1 */

    B_tmp = CreateBTemp (rw->num_labels, rw->num_unlabeled);
    if (B_tmp == NULL)
        return 1;

    /* Traversing all the unlabeled voxels */
    for (z = p = 0; z < rw->input->zsize; ++z) {
        for (y = 0; y < rw->input->ysize; ++y) {
            for (x = 0; x < rw->input->zsize; ++x, ++p) {
                p_index = rw->labels->data[p];
                if (UNLABELED(p_index)) {
                    SetMatrixValue (T, p_index, p_index, 0);
                    /* Traversing all the neighbouring voxels of the
                     * current voxel */
                    for(i = 0; i < rw->adj->n; ++i){
                        adj_x = x + rw->adj->dx[i];
                        adj_y = y + rw->adj->dy[i];
                        adj_z = z + rw->adj->dz[i];
                        if ((adj_x == x) && (adj_y == y) && (adj_z == z))
                            continue;

                        if (ValidVoxel (rw->labels, adj_x, adj_y, adj_z)){
                            q =   adj_x
                                + adj_y * rw->labels->xsize
                                + adj_z * rw->labels->xsize * rw->labels->ysize;
                            q_index = rw->labels->data[q];
                            /* Filling the temporary T and B_tmp matrixes */
                            if (UNLABELED(q_index)) {
                                if (q_index > p_index) {
                                    weight = GetWeight (rw->input, p, q);
                                    SetMatrixValue (T, p_index, q_index, -weight);
                                    AddMatrixValue (T, p_index, p_index, weight);
                                } else {
                                    /* We can use the previously calculated weight, instead of calculating it again */
                                    CopyNewMatrixValue (T, q_index, p_index, p_index, q_index, &weight);
                                    AddMatrixValue (T, p_index, p_index, -weight);
                                }
                            } else {
                                q_index = -q_index - 1;
                                weight = GetWeight (rw->input, p, q);
                                AddMatrixValue (T, p_index, p_index, weight);
                                B_tmp[q_index][p_index] += weight;
                            }
                        }
                    }
                }
            }
        }
    }
    DestroyScene (&(rw->input));
    /* Passing the B_tmp values to the final B matrix */
    for (i = 0; i < rw->num_labels; ++i)
        for (j = 0; j < rw->num_unlabeled; ++j)
            lis_vector_set_value (LIS_INS_VALUE, j, B_tmp[i][j], rw->lin_sys->B[i]);
    DestroyBTemp (&B_tmp, rw->num_labels);
    /* Transforming the T matrix to the correct format,
     * assignint it to the A matrix */
    rw->lin_sys->A = GetLisMatrix (T);
    DestroyTempSparseMatrix (&T);
    return 0;
}

/*
 * Creates the linear systems to be solved,
 * allocating memory and filling the matrixes values.
 */
static int CreateLinearSystem (RandomWalk rw) {
    LIS_INT    A_n = rw->num_unlabeled;
    LIS_INT    x_n = A_n;
    LIS_INT    b_n = A_n;
    int l;

    rw->lin_sys->B = (LIS_VECTOR *) malloc (sizeof(*(rw->lin_sys->B)) * rw->num_labels);
    if (rw->lin_sys->B == NULL) {
        fprintf (stderr, "Error allocating B matrix\n");
        return 1;
    }
    for (l = 0; l < rw->num_labels; ++l) {
        lis_vector_create (0, rw->lin_sys->B + l);
        lis_vector_set_size (rw->lin_sys->B[l], 0, b_n);
    }

    lis_vector_create (0, &(rw->lin_sys->x));
    lis_vector_set_size (rw->lin_sys->x, 0, x_n);
    lis_vector_set_all (0, rw->lin_sys->x);

    rw->last_x = (double *) malloc (sizeof(*(rw->last_x)) * x_n);
    if (rw->last_x == NULL) {
        fprintf (stderr, "Error allocating last_x vector\n");
        return 1;
    }
    for (l = 0; l < x_n; ++l)
        rw->last_x[l] = 1.;

    rw->max_prob = (double *) malloc (sizeof(*(rw->max_prob)) * x_n);
    if (rw->max_prob == NULL) {
        fprintf (stderr, "Error allocating max_prob vector\n");
        return 1;
    }
    for (l = 0; l < x_n; ++l)
        rw->max_prob[l] = 0.;

    if (FillAAndB (rw))
        return 2;

    return 0;
}

/*
 * Frees up the memory allocated by CreateLinearSystem
 */
static int DestroyLinearSystem (LIS_MATRIX A, LIS_VECTOR x, LIS_VECTOR **B,
                                double **last_x, double **max_prob, int num_labels) {
    int l;
    for (l = 0; l < num_labels; ++l) {
        lis_vector_destroy ((*B)[l]);
    }
    free(*B);
    *B = NULL;
    lis_vector_destroy (x);
    lis_matrix_destroy (A);
    free (*last_x);
    *last_x = NULL;
    free (*max_prob);
    *max_prob = NULL;
    return 0;
}

/*
 * Returns a sequenced scene file name for the labels probabilities.
 * This file name is based on the string output_scene_file_ and the number label_num.
 */
static char * CreateLabelSceneFileName (char *output_scene_file_, int label_num) {
    char *scene_file_name = NULL;
    int scene_file_name_size, n, m;
    char buffer[50];
    m = strlen(output_scene_file_);
    if (m >= 4 && strcmp (output_scene_file_ + m - 4, ".scn") == 0)
        m -= 4;
    n = sprintf (buffer, "_prob_label_%d.scn", label_num);
    scene_file_name_size = m + n + 1;
    scene_file_name = (char *) malloc (sizeof(char) * scene_file_name_size);
    if (scene_file_name == NULL) {
        fprintf (stderr, 
                "Error allocating memory for label scene of the label %d\n",
                label_num);
        return NULL;
    }
    strncpy (scene_file_name, output_scene_file_, m);
    scene_file_name[m] = '\0';
    strncat (scene_file_name, buffer, n);
    return scene_file_name;
}

/*
 * Writes the label probabilities scene file, and also the final
 * segmentation if it is the last label being written.
 */
static void WriteLabelScene (char *output_scene_file, Scene *output,
                             Scene *labels, int label_num, LIS_VECTOR x,
                             double *last_x, double *max_prob, int is_last_label,
                             int maxval) {
    int k, voxel_index;
    double value;
    Scene *scene = NULL;
    char *scene_file_name = NULL;
    scene_file_name = CreateLabelSceneFileName (output_scene_file, label_num);
    scene = CreateScene (labels->xsize, labels->ysize, labels->zsize);
    if (scene_file_name == NULL || scene == NULL) {
        fprintf (stderr,
                "Error creating label scene for label %d\n", label_num);
        return;
    }
    SetSceneImax (scene, maxval);
    for (k = 0; k < labels->n; ++k) {
        if (labels->data[k] >= 0) {
            voxel_index = labels->data[k];
            if (!is_last_label) {
                lis_vector_get_value (x, voxel_index, &value);
                last_x[labels->data[k]] -= value;
            } else {
                value = last_x[voxel_index];
            }
            if (value > max_prob[voxel_index]) {
                max_prob[voxel_index] = value;
                output->data[k] = label_num - 1;
            }
            scene->data[k] = (int) (value * maxval + 0.5);
            if (scene->data[k] < 0) {
                fprintf (stderr, "Found negative probability!\n");
                scene->data[k] = 0;
            } else if (scene->data[k] > maxval) {
                fprintf (stderr, "Found probability bigger than 1!\n");
                scene->data[k] = maxval;
            }
        } else {
            /* This voxel is already labeled by the input */
            if (labels->data[k] == -label_num) {
                scene->data[k] = maxval;
                output->data[k] = label_num - 1;
            } else {
                scene->data[k] = 0;
            }
        }
    }
    WriteScene (scene, scene_file_name);
    DestroyScene (&scene);
    free (scene_file_name);
    scene_file_name = NULL;
    if (is_last_label) {
        WriteScene (output, output_scene_file);
    }
}

/*
 * Solves the linear system and writes the results
 * to label probabilities scenes files.
 * Also writes the final segmentation, with the assigned label indexes,
 * on the output_scene_file.
 */
static void Solve (RandomWalk rw) {
    LIS_SOLVER solver;
    LIS_INT iter;
    int l;

    lis_solver_create(&solver);
    lis_solver_set_optionC(solver);
    lis_solver_set_option("-i cg -p ssor", solver);
    lis_solver_set_option("-tol 1.0e-12 -maxiter 10000", solver);
    for (l = 0; l < rw->num_labels - 1; ++l) {
        lis_solve (rw->lin_sys->A, rw->lin_sys->B[l], rw->lin_sys->x, solver);
        lis_solver_get_iters (solver, &iter);
        printf ("Number of iterations to solve linear system: %d\n", iter);
        WriteLabelScene (rw->output_scene_file, rw->output, rw->labels, l + 1,
                rw->lin_sys->x, rw->last_x, rw->max_prob, 0, rw->maxval);
    }
    WriteLabelScene (rw->output_scene_file, rw->output, rw->labels, l + 1,
            rw->lin_sys->x, rw->last_x, rw->max_prob, 1, rw->maxval);
    lis_solver_destroy(solver);
}

/* Executes the Random Walks algorithm with the information stored on rw */
int RANDOMWALKrun (RandomWalk rw) {
    if (   rw->input->xsize != rw->labels->xsize
        || rw->input->ysize != rw->labels->ysize
        || rw->input->zsize != rw->labels->zsize ) {
        fprintf (stderr, "Input scene size and label size differ!\n");
        return 1;
    }
    if (IndexVoxels (rw->labels, &(rw->num_labels), &(rw->num_unlabeled)))
        return 2;
    if (CreateLinearSystem (rw))
        return 3;
    printf ("Unlabeled voxels: %d\n", rw->num_unlabeled);
    Solve (rw);
    DestroyLinearSystem (rw->lin_sys->A, rw->lin_sys->x, &(rw->lin_sys->B),
            &(rw->last_x), &(rw->max_prob), rw->num_labels);
    return 0;
}
